/* bob.c - the main routine */
/*
	Copyright (c) 1991, by David Michael Betz
    BOB+ Copyright (c) 2005-2005 by R,-Erik Ebert
	All rights reserved
*/

#include <stdio.h>
#include <setjmp.h>
#include "bob.h"

#ifndef BP_RUN
  #define USE_COMPILE
#endif

static char* BANNER	=
#ifndef BP_RUN
   #ifdef BP_DEMO
      "Bob+ v1.1a Demo - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
      "Parts are based on code of Bob, Copyright (c) 1991, by David M. Betz\n\n"
      "usage:\n"
      "bpdemo [-i][-d][-t] [sourcefile [..]] [# userarg [..]]\n"
      "options:\n"
      "    -d ... debug information during compilation\n"
      "    -t ... trace bytecodes during execution\n"
      "    -i ... info (shows banner, syntax and names of compiled functions)\n"
      "     # ... marks begin of user argumentlist\n";
   #else
      #ifdef BP_COMP
         "Bob+ v1.1a Compiler - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
         "Parts are based on code of Bob, Copyright (c) 1991, by David M. Betz\n\n"
         "usage:\n"
         "bpc [-i][-d] [sourcefile [..]] [-r objfile [..]] [-o outfile]\n"
         "options:\n"
         "    -d ... debug information during compilation\n"
         "    -i ... info (shows banner, syntax and names of compiled functions)\n"
         "    -r ... specifies precompiled files to be loaded\n"
         "    -o ... name of output file (only used with -c)\n";
      #else
         #ifdef USE_FLOAT
            "Bob+ v1.1a - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
         #else
            "Bob+ v1.1ai - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
         #endif
         "Parts are based on code of Bob, Copyright (c) 1991, by David Betz\n\n"
         "usage:\n"
         #ifdef USE_FLOAT
            "bp [-i][-d][-t][-c] [sourcefile [..]] [-r objfile [..]] \n"
         #else
            "bpi [-i][-d][-t][-c] [sourcefile [..]] [-r objfile [..]] \n"
         #endif
         "   [-o outfile] [# userarg [..]]\n"
         "options:\n"
         "    -d ... debug information during compilation\n"
         "    -t ... trace bytecodes during execution\n"
         "    -i ... info (shows banner, syntax and names of compiled functions)\n"
         "    -c ... compile only\n"
         "    -r ... specifies precompiled files to be loaded\n"
         "    -o ... name of output file (only used with -c)\n"
         "     # ... marks begin of user argumentlist\n";
      #endif  /* ! BP_COMP */
   #endif /* ! DEMO */
#else  /* BP_RUN */
   #ifdef USE_FLOAT
      "Bob+ runtime v1.1a - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
   #else
      "Bob+ runtime v1.1ai - Copyright (c) 2005-2008, by Ralf-Erik Ebert\n"
   #endif
      "Parts are based on code of Bob, Copyright (c) 1991, by David M. Betz\n\n"
      "usage:\n"
   #ifdef USE_FLOAT
      "bpr [-i][objfile [..]] [# userarg [..]]\n"
   #else
      "bpri [-i][objfile [..]] [# userarg [..]]\n"
   #endif
   "options:\n"
   "    -i ... info (shows banner and syntax)\n"
   "     # ... marks begin of user argumentlist\n";
#endif
/* global variables */

jmp_buf error_trap;

/* external variables */
extern int decode,trace;
extern int usrarg0;
extern int cmdargc;
extern char** cmdargv;
extern VALUE* sp;

int verbose = 0;
int compile = 0;
int load = 0;
char* outname = "out.bpm";

/*
*/
/* main - the main routine */
main(argc,argv)
  int argc; char *argv[];
{
	char* bpstack;
    char fullname[20];
    int i;
    int firstsrc = 0, srccnt = 0;
    /* initialize */
    bpstack = getenv("BPSTACK");
    i = 0;
    if (bpstack != 0)
    {
		i = atoi(bpstack);
	}
	if (i == 0) i = SMAX;
    initialize(i,CMAX);
    usrarg0 = argc;
    cmdargc = argc;
    cmdargv = argv;
    /* load and execute some code */
    if (argc == 1) verbose = 1;
    for (i = 1; i < argc; ++i)
    if (strcmp(argv[i],"#") == 0)
    {
		usrarg0 = i+1;
		break;
	}
	else if (argv[i][0] == '-')
	{
		switch (argv[i][1])
		{
#ifndef BP_RUN
		case 'd':
		   decode = 1;
		   break;
#endif
		case 'i':
		   verbose = 1;
		   break;
#ifndef BP_RUN
	#ifndef BP_COMP
		case 't':
		   trace = 1;
		   break;
		case 'c':
		#ifdef BP_DEMO
           printf("compile option not supported in demo version\n");
           return 1;
		#else
		   compile = 1;
		   break;
		#endif /* ! BP_DEMO */
	#endif	/* !BP_COMP */
	#ifndef BP_DEMO
		case 'o':
		   if (i<argc-1)
		   {
			   i++;
			   outname = argv[i];
		   }
		   break;
    #endif
		case 'r':
#else
        default:
#endif // BP_RUN
#ifdef BP_DEMO
           printf("read module option not supported in demo version\n");
           return 1;
#else // ! DEMO
		   while (i < argc-1)
		   {
			   if ( (argv[i+1][0] == '-') || (argv[i+1][0] == '#'))
			     break;
			   strcpy(fullname,argv[i+1]);
 		       if (strchr(fullname,'.') == 0)
				  strcat(fullname,".bpm");
			   if (!readModule(fullname))
			   {
				   printf("error reading module %s\n",fullname);
				   return 1;
			   }
			   i++;
		   }
#endif // !BP_DEMO
#ifndef BP_RUN
         break;
		default:
		   printf("unknown option %s ignored\n",argv[i]);
#endif
		}
	}
	else
	{
		if (srccnt == 0) firstsrc = i;
		srccnt++;
	}
#ifdef BP_COMP
       if (srccnt > 0) compile=1;
#endif
	for (i=firstsrc;i < firstsrc+srccnt;i++)
	{
#ifndef BP_RUN
	    strcpy(fullname,argv[i]);
		if (strchr(fullname,'.') == 0)
		   strcat(fullname,".bp");
		if(!compile_file(fullname))
		   return 255;
#else
	   strcpy(fullname,argv[i]);
	   if (strchr(fullname,'.') == 0)
		  strcat(fullname,".bpm");
	   if (!readModule(fullname))
	   {
		   printf("error reading module %s\n",fullname);
		   return 1;
	   }
#endif
	}
	if (verbose)
	     osputs(BANNER);
#ifdef BP_DEMO
     else osputs("BOB+ 1.0 demo version\n");
#endif
	if (compile)
	{
#ifndef BP_DEMO
	   i = (writeModule(outname)) ? 0 : 1;
#endif
    }
#ifndef BP_COMP
	else
	{
       if (!execute("main"))
          return 1;
       i = 0;
       if (sp->v_type == DT_INTEGER)
          i = sp->v.v_integer;
    }
#endif
    return i;
}

#ifdef USE_COMPILE
/* compile_file - compile definitions in a file */
int compile_file(name)
  char *name;
{
    FILE *ifp;
    int result = FALSE;
    if ((ifp = fopen(name,"r")) != NULL) {
	result = compile_definitions(fgetc,ifp);
	fclose(ifp);
    }
    return result;
}
#endif

/* info - display progress information */
info(fmt,a1,a2,a3,a4,a5,a6)
  char *fmt;
{
    if (!(decode || trace || verbose))  return;
	osputs("[ ");
    printf(fmt,a1,a2,a3,a4,a5,a6);
	osputs(" ]\n");
 /*   char buf1[100],buf2[100];

    sprintf(buf1,fmt,a1,a2,a3,a4,a5,a6);
    sprintf(buf2,"[ %s ]\n",buf1);
    osputs(buf2); */
}

/* error - print an error message and exit */
error(fmt,a1,a2,a3,a4)
  char *fmt;
{
	osputs("Error: ");
	printf(fmt,a1,a2,a3,a4);
	osputs("\n");
/*    char buf1[300],buf2[300];
    sprintf(buf1,fmt,a1,a2,a3,a4);
    sprintf(buf2,"Error: %s\n",buf1);
    osputs(buf2); */
    longjmp(error_trap,1);
}

osputc(ch)
  int ch;
{
    putc(ch,stdout);
}

osputs(str)
  char *str;
{
    fputs(str,stdout);
}
