#include "bob.h"
#include <string.h>
#include <setjmp.h>
#ifndef BP_DEMO
extern jmp_buf error_trap;	/* trap for compile errors */

extern DICTIONARY *symbols;	/* symbol table */
extern DICTIONARY *classes;	/* bpclass table */
extern DICT_ENTRY* sym_end;

static int writeBytes(FILE* fp, void* p, int cnt)
{
	if (fwrite(p,cnt,1,fp) > 0)
	   return cnt;
	return 0;
}

static int writeString(FILE* fp, char* str)
{
	if (str == 0)
	   return writeString(fp,"");
	return writeBytes(fp, str, strlen(str)+1);
}


static char* findFunctionNameFromPointer(void* ptr)
{
	DICT_ENTRY* p;
   int type;
	for (p=symbols->di_contents; p!= 0; p=p->de_next)
	{
      type = p->de_value.v_type;
		if ((type == DT_CODE) || (type == DT_FILE))
		{
			if (p->de_value.v.v_code == ptr)
			   return p->de_key;
		}
	}
	return 0;
}

int writeDictEntry();
int writeDict();
char* getDictName();
DICTIONARY* getDictFromName();


typedef struct objlistentry
{
	int ID;
	VALUE* val;
	int type;
	struct objlistentry* next;
} OL_ENTRY;

static OL_ENTRY* objlist = 0;
static OL_ENTRY* oreflist = 0;

static int addOLEntry(int ID, VALUE* val, int type)
{
	OL_ENTRY* ep;
	ep = objlist;
	if (ID == 0)
	   ID = (int) val;
	while (ep != 0)
	{
		if (ep->ID == ID)
		   return 0;
		ep = ep->next;
	}
	ep = (OL_ENTRY*) malloc(sizeof(OL_ENTRY));
	ep->ID = ID;
	ep->type=type;
	ep->val = val;
	ep->next = objlist;
	objlist = ep;
	return 1;
}

static void addORefEntry(int ID, VALUE* val)
{
	OL_ENTRY* ep;
	ep = oreflist;
	ep = (OL_ENTRY*) malloc(sizeof(OL_ENTRY));
	ep->ID = ID;
	ep->val = val;
	ep->type=DT_NIL;
	ep->next = oreflist;
	objlist = ep;
}

static void clearobjectlists()
{
	OL_ENTRY* ep;
	while (objlist != 0)
	{
		ep = objlist;
		objlist = ep->next;
		free(ep);
	}
	while (oreflist != 0)
	{
		ep = oreflist;
		oreflist = ep->next;
		free(ep);
	}
}

static void dissolveObjRefs()
{
	OL_ENTRY *po, *pr;
	while (objlist != 0)
	{
		po = objlist;
		objlist = po->next;
		pr = oreflist;
		while (pr != 0)
		{
			if (pr->ID == po->ID)
			{
			   pr->val->v_type = po->type;
			   pr->val->v.v_object = po->val;
		    }
		    pr = pr->next;
		}
		objlist = po->next;
		free(po);
	}
	while (oreflist != 0)
	{
		pr = oreflist->next;
		free(oreflist);
		oreflist = pr;
	}
}

static int writeValue(FILE* fp, VALUE* src)
{
	long i,n;
   static int writeclassref = 0;
	if (src == 0) return 0;
	writeBytes(fp,&(src->v_type),sizeof(int));
	switch (src->v_type)
	{
		case DT_CLASS:
		   if (src->v.v_class == 0) return 0;
		   writeString(fp,src->v.v_class->cl_name);
         if (!writeclassref)
         {
		     if (src->v.v_class->cl_base == 0)
		        writeString(fp,"");
		     else
		        writeString(fp,src->v.v_class->cl_base->cl_name);
		     writeDict(fp,src->v.v_class->cl_members);
		     writeDict(fp,src->v.v_class->cl_functions);
		     writeBytes(fp,&(src->v.v_class->cl_size),sizeof(int));
         }
		   break;
		case DT_OBJECT:
		   if (src->v.v_object == 0) return 0;
		   writeBytes(fp,&(src->v.v_object),sizeof(int)); /* ID */
		   if (addOLEntry(0,src->v.v_object,DT_OBJECT))
		   {
		      writeString(fp,src->v.v_object[0].v.v_class->cl_name);
		      n=src->v.v_object[0].v.v_class->cl_size;
		      for (i=1; i<=n; i++)
		         writeValue(fp,&(src->v.v_object[i]));
	       }
	       else
		      writeString(fp,"");
		   break;
		case DT_STRING:
		   if (src->v.v_string == 0) return 0;
		   writeBytes(fp,&(src->v.v_string->s_length),sizeof(int));
		   writeBytes(fp,src->v.v_string->s_data,src->v.v_string->s_length);
		   break;
		case DT_BYTECODE:
     	case DT_VECTOR:
		   if (src->v.v_vector == 0)
         {
            return 0;
         }
		   writeBytes(fp,&(src->v.v_vector),sizeof(int)); /* ID */
		   if (addOLEntry(0,src->v.v_object,DT_VECTOR))
		   {
			   n = src->v.v_vector[0].v.v_integer;
			   writeBytes(fp,&n,sizeof(long));
			   for (i=1; i<=n; i++)
			   {
				  writeclassref = 1;
				  writeValue(fp,&(src->v.v_vector[i]));
			   }
			   writeclassref = 0;
           }
           else
           {
              n = -1;
		      writeBytes(fp,&n,sizeof(long));
	       }
		   break;
		case DT_CODE:
		case DT_FILE:
		   writeString(fp,findFunctionNameFromPointer(src->v.v_code));
		   break;
		case DT_VAR:
		    writeString(fp,getDictName(src->v.v_var));
          writeString(fp,src->v.v_var->de_key);
		   /* writeDictEntry(fp,src->v.v_var); */
			break;
		case DT_INTEGER:
		case DT_FLOAT:
			writeBytes(fp,&(src->v),sizeof(float));
	}
	return 1;
}

static char* getDictName(DICT_ENTRY* entry)
{
	static char buf[82];
	CLASS* pc;
	if (entry == 0)
	   return 0;
	if (entry->de_dictionary == 0)
	   return 0;
	if (entry->de_dictionary == symbols)
	   return "__S";
	if (entry->de_dictionary == classes)
	   return "__C";
	pc = entry->de_dictionary->di_class;
	if (pc == 0)
	   return 0;
	sprintf(buf,"%s_", pc->cl_name);
	if (entry->de_dictionary == pc->cl_members)
	   strcat(buf,"m");
	else
	   strcat(buf,"f");
	return buf;
}

static DICTIONARY* getDictFromName(char* name)
{
	int n;
	int ismember = 0;
	DICT_ENTRY* entry;
	CLASS* pc;
	if (strcmp(name ,"__S")==0)
	   return symbols;
	if (strcmp(name,"__C") == 0)
	   return classes;
	n = strlen(name);
	if (n < 3) return 0;
	if (name[n-1] == 'm') ismember = 1;
	name[n-2] = '\0';
	entry = findentry(classes,name);
	if (entry == 0) return 0;
	pc = entry->de_value.v.v_class;
	if (pc == 0)
	   return 0;
	if (ismember)
	   return pc->cl_members;
	return pc->cl_functions;
}

static int writeDictEntry(FILE* fp, DICT_ENTRY* src)
{
	if (src == 0) return 0;
	/*writeString(fp,getDictName(src)); */
	writeString(fp,src->de_key);
	writeBytes(fp,&(src->de_type),sizeof(int));
	writeValue(fp,&(src->de_value));
	return 1;
}

static int writeDict(FILE* fp, DICTIONARY* src)
{
	DICT_ENTRY *p, *endp;
	int cnt;
	DICT_ENTRY** pe;
	if (src == 0) return 0;
	endp = (src == symbols) ? sym_end : 0;
	if (src->di_class != 0)
	   writeString(fp,src->di_class->cl_name);
	else
	   writeString(fp,"");
	for (cnt=0,p=src->di_contents; p != endp; cnt++, p = p->de_next);
	if (cnt > 0)
	{
		pe = (DICT_ENTRY**) calloc(cnt,sizeof(DICT_ENTRY*));
		for (cnt=0,p=src->di_contents; p != endp; cnt++, p = p->de_next)
		   pe[cnt] = p;
		for (cnt--;cnt >= 0;cnt--)
		{
			if (!writeDictEntry(fp,pe[cnt]))
			{
			   free(pe);
			   return 0;
		    }
		}
		free(pe);
	}
	writeString(fp,""); /* write zero byte as end sign */
	return 1;
}

static int readBytes(FILE* fp, void* p, int cnt)
{
	return fread(p,cnt,1,fp) ;
}

static char* readString(FILE* fp)
{
	char* p;
	static char buf[200];
	buf[0] = 0;
	for (p=buf;fread(p,1,1,fp) > 0 ;p++)
	{
		if (*p == '\0') break;
	}
	*p = '\0';
	return buf;
}

typedef struct varlistentry
{
	VALUE* var;
	char* dictname;
	char* key;
	struct varlistentry* next;
} VL_ENTRY;

static VL_ENTRY* varlist = 0;

static void readVlEntry(VALUE* dest, FILE* fp)
{
   VL_ENTRY* vp = (VL_ENTRY*) malloc(sizeof(VL_ENTRY));
   vp->var = dest;
   vp->dictname = strdup(readString(fp));
   vp->key = strdup(readString(fp));
   vp->next = varlist;
   varlist = vp;
}

static void dissolveVars()
{
	DICTIONARY* dict;
	DICT_ENTRY* entry;
	VL_ENTRY* vp;
	while (varlist != 0)
	{
		vp = varlist;
        dict = getDictFromName(vp->dictname);
        if (dict != 0)
        {
			entry = findentry(dict,vp->key);
			if (entry != 0)
			   vp->var->v.v_var=entry ;
		}
		free(vp->dictname);
		free(vp->key);
		varlist = vp->next;
		free(vp);
	}
}

static int readValue(FILE* fp, VALUE* dest)
{
	char buf[200];
	char* p;
	long i,n;
	int l;
   static int readclassref = 0;
	CLASS* base = 0;
	DICT_ENTRY* entry;
	DICTIONARY* dict;
	int type;
	if (dest == 0) return 0;
	set_nil(dest);
	readBytes(fp,&(dest->v_type),sizeof(int));
	switch (dest->v_type)
	{
		case DT_CLASS:
		   strcpy(buf,readString(fp)); /* classname */
         if (!readclassref)
         {
		     p = readString(fp); /* name of baseclass */
		     if (strlen(p) > 0)
		     {
		         entry = findentry(classes,p);
		         if (entry != 0)
		            base = entry->de_value.v.v_class;
		     }
		     dest->v.v_class=newclass(buf,base);
		     readDict(fp,dest->v.v_class->cl_members);
		     readDict(fp,dest->v.v_class->cl_functions);
		     readBytes(fp,&(dest->v.v_class->cl_size),sizeof(int));
         }
         else
         {
            entry = findentry(classes,buf);
            if (entry != 0)
               dest->v.v_class = entry->de_value.v.v_class;
         }
		   break;
		case DT_OBJECT:
		   readBytes(fp,&l,sizeof(int)); /* ID */
		   p = readString(fp); /* classname */
		   if (strlen(p) > 0)
		   {
		       entry = findentry(classes,p);
		       if (entry != 0)
		          base = entry->de_value.v.v_class;
  		      if (base == 0) return 0;
		      dest->v.v_object = newobject(base);
		      addOLEntry(l,dest->v.v_object,DT_OBJECT);
		      n = base->cl_size;
		      for (i=1;i<=n;i++)
		         readValue(fp,&(dest->v.v_object[i]));
		   }
		   else
		   {
		      addORefEntry(l,dest);
		      set_nil(dest);
	       }
		   break;
		case DT_STRING:
		   /*if (dest->v.v_string == 0) return 0; */
		   readBytes(fp,&l,sizeof(int));
		   dest->v.v_string = newstring(l);
		   /*readBytes(fp,&(dest->v.v_string->s_length),sizeof(int));*/
		   readBytes(fp,dest->v.v_string->s_data,l);
		   break;
		case DT_VECTOR:
		case DT_BYTECODE:
		   readBytes(fp,&l,sizeof(int)); /* ID */
		   readBytes(fp,&n,sizeof(long));
		   if (n>=0)
		   {
			   dest->v.v_vector = newvector(n);
		       addOLEntry(l,dest->v.v_vector,DT_VECTOR);
			   for (i=1; i<=n; i++)
			   {
				  readclassref = 1;
				  readValue(fp,&(dest->v.v_vector[i]));
			   }
         	   readclassref = 0;
	       }
	       else
		   {
		      addORefEntry(l,dest);
		      set_nil(dest);
	       }
		   break;
		case DT_CODE:
		case DT_FILE:
		   p = readString(fp);
		   entry = findentry(symbols,p);
		   if (entry == 0) return 0;
		   dest->v.v_code = entry->de_value.v.v_code;
		   break;
		case DT_VAR:
		    readVlEntry(dest,fp);
			break;
		case DT_INTEGER:
		case DT_FLOAT:
			readBytes(fp,&(dest->v),sizeof(float));
	}
	return 1;
}

static int readDictEntry(FILE* fp, DICTIONARY* dest)
{
	char* key;
	int type;
    DICT_ENTRY* entry;
    if (dest == 0) return 0;
    key = readString(fp);
    if (strlen(key) == 0)
       return 0;
    readBytes(fp,&type,sizeof(int));
    entry = addentry(dest,key,type);
    return readValue(fp,&(entry->de_value));
}


static int readDict(FILE* fp, DICTIONARY* dest)
{
	char* clname;
	if (dest == 0)
	   return 0;
	clname = readString(fp);
/*	if (strlen(clname) != 0)
	{
	} */
	while (readDictEntry(fp, dest));

	return 1;
}

int readModule(char* filename)
{
	FILE* fp;
	int result = 0;
	varlist = 0;
	objlist = 0;
	oreflist = 0;
    /* trap errors */
    if (setjmp(error_trap))
	return (FALSE);

	fp = fopen(filename,"rb");
	if (fp != 0)
	{
		if (result = readDict(fp,classes))
		   result = readDict(fp,symbols);
		fclose(fp);
	}
	dissolveVars();
	dissolveObjRefs();
	return result;
}

int readSingleValue(FILE* fp, VALUE* value)
{	int result = 0;
	if ((fp == 0) || (value == 0)) return result;
	varlist = 0;
	objlist = 0;
	oreflist = 0;
    result = readValue(fp,value);
	dissolveVars();
	dissolveObjRefs();
	return result;
}

int writeSingleValue(FILE* fp, VALUE* value)
{
	int result;
	if ((fp == 0) || (value == 0)) return 0;
	objlist = 0;
	oreflist = 0;
	result =  writeValue(fp,value);
	clearobjectlists();
	return result;
}


int writeModule(char* filename)
{
	FILE* fp;
	int result = 0;
    /* trap errors */
    if (setjmp(error_trap))
	return (FALSE);
	objlist = 0;
	oreflist = 0;

	fp = fopen(filename,"wb");
	if (fp != 0)
	{
		if (result = writeDict(fp,classes))
		   result = writeDict(fp,symbols);
		fclose(fp);
	}
	clearobjectlists();
	return result;
}
#endif // !BP_DEMO